/* 
 *  ENTER emulator (c) Copyright, Kevin Thacker 1999-2001
 *  
 *  This file is part of the ENTER emulator source code distribution.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* PC CMOS RTC */
#include "rtc.h"
/*
%	Reg#			  Description

	 00  RTC seconds
	 01  RTC seconds alarm
	 02  RTC minutes
	 03  RTC minutes alarm
	 04  RTC hours
	 05  RTC hours alarm
	 06  RTC day of week
	 07  RTC day of month
	 08  RTC month
	 09  RTC year
	 0A  RTC Status register A:

	    76543210  RTC Status Register A
	          rate selection Bits for divider output
	        		 frequency (set to 0110 = 1.024kHz, 976.562s)
	       22 stage divider, time base being used;
	     			  (initialized to 010 = 32.768kHz)
	      1=time update in progress, 0=time/date available
	 
	 0B  RTC Status register B:

	    76543210  RTC Status Register B
	             1=enable daylight savings, 0=disable (default)
	            1=24 hour mode, 0=12 hour mode (24 default)
	           1=time/date in binary, 0=BCD (BCD default)
	          1=enable square wave frequency, 0=disable
	         1=enable update ended interrupt, 0=disable
	        1=enable alarm interrupt, 0=disable
	       1=enable periodic interrupt, 0=disable
	      1=disable clock update, 0=update count normally

	 0C  RTC Status register C (read only):

	    76543210  RTC Status Register C  (read only)
	          reserved (set to 0)
	         update ended interrupt enabled
	        alarm interrupt enabled
	       periodic interrupt enabled
	      IRQF flag

	 0D  RTC Status register D (read only):

	    76-0  RTC Status Register D  (read only)
	     	 reserved (set to 0)
	      1=CMOS RAM has power, 0=CMOS RAM has lost power
*/

typedef struct CMOS
{
	unsigned long RegisterIndex;
	unsigned char Registers[64];
} CMOS;


static CMOS cmos;


enum
{
	RTC_REGISTER_SECONDS = 0,
	RTC_REGISTER_SECONDS_ALARM,
	RTC_REGISTER_MINUTES,
	RTC_REGISTER_MINUTES_ALARM,
	RTC_REGISTER_HOURS,
	RTC_REGISTER_HOURS_ALARM,
	RTC_REGISTER_DAY_OF_WEEK,
	RTC_REGISTER_DAY_OF_MONTH,
	RTC_REGISTER_MONTH,
	RTC_REGISTER_YEAR,
	RTC_STATUS_REGISTER_A,
	RTC_STATUS_REGISTER_B,
	RTC_STATUS_REGISTER_C
} RTC_REGISTER_ID;


static void CMOS_Update()
{



}


void	CMOS_SelectRegister(unsigned char RegisterIndex)
{
	cmos.RegisterIndex = RegisterIndex & 0x03f;
}

void	CMOS_WriteData(unsigned char Data)
{
	cmos.Registers[cmos.RegisterIndex] = Data;

	cmos.Registers[RTC_REGISTER_SECONDS] = 5;
	cmos.Registers[RTC_REGISTER_MINUTES] = 24;
	cmos.Registers[RTC_REGISTER_HOURS] = 12;
	cmos.Registers[RTC_REGISTER_DAY_OF_WEEK] = 7;
	cmos.Registers[RTC_REGISTER_DAY_OF_MONTH] = 26;
	cmos.Registers[RTC_REGISTER_MONTH] = 12;
	cmos.Registers[RTC_REGISTER_YEAR] = 99;
}

unsigned char CMOS_ReadData(void)
{
	return cmos.Registers[cmos.RegisterIndex];
}
